package genetic;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Random;

import board.Board;
import board.DraughtPiece;
import gameControl.GameState;
import moveFinders.LegalPieceFinder;
import players.HumanPlayer;
import players.RandomPlayer;
import tree.ChildNodeFinder;
import tree.MiniMaxNode;

public class GeneticPlayer {
	private int playerNum;
	private GameState gameState;
	private int opponentNum;
	
	private Board board;
	
	private int[] heuristicWeights;
	
	private GeneticPlayer geneticOpponent;
	private RandomPlayer randomOpponent;
	private String opponentType;
	
	private int totalTurnsTaken;
	
	
	public GeneticPlayer(int playerNum, GameState gameState, int[] heuristicWeights, Board board) {
		this.playerNum=playerNum;
		this.gameState=gameState;
		
		if(playerNum==1) {
			opponentNum=2;
		} 
		if(playerNum==2) {
			opponentNum=1;
		}
		this.board=board;
		this.heuristicWeights=heuristicWeights;
		
		totalTurnsTaken=0;
	}
	
	

	public void start() {
		++totalTurnsTaken;
	
		if(totalTurnsTaken<100) {
			LegalPieceFinder legalPieceFinder=new LegalPieceFinder();
			ArrayList<DraughtPiece> legalPieces=legalPieceFinder.findLegalPieces(playerNum, gameState.getGameStatePieces());
			if(!legalPieces.isEmpty()) {
				MiniMaxNode rootNode=new MiniMaxNode(gameState.getGameStatePieces(), 0.0, "maximiser");
				simulateGame(rootNode);
				doMove(rootNode);
			}
		} 
		
		
		
		
		
	}

	
	
	
	
	private void doMove(MiniMaxNode rootNode) {
		/*ArrayList<MiniMaxNode> levelOneNodes=rootNode.getChildren();
		MiniMaxNode levelTwoNode=levelOneNodes.get(0);
		ArrayList<MiniMaxNode> levelThreeNodes=levelTwoNode.getChildren();
		MiniMaxNode levelThreeNode=levelThreeNodes.get(0);
		ArrayList<MiniMaxNode> leafNodes=levelThreeNode.getChildren();
		MiniMaxNode leafNode=leafNodes.get(0);
		*/
		
		miniMax(rootNode);
		
		ArrayList<MiniMaxNode> possibleNodes=new ArrayList<MiniMaxNode>();
		for(MiniMaxNode node:rootNode.getChildren()) {
			if(node.getMiniMaxScore().equals(rootNode.getMiniMaxScore())) {
				possibleNodes.add(node);
			}
		}
		
		if(possibleNodes.size()>0) {
			Random randGenerator=new Random();
			int randNum=randGenerator.nextInt(possibleNodes.size());
			
			MiniMaxNode chosenNode=possibleNodes.get(randNum);
			ArrayList<DraughtPiece> chosenMove=chosenNode.getGameStatePieces();
			

			gameState.setGameState(chosenMove);
		
			end();
		} 
		
	}



	private void end() {
		board.repaintBoard(null, gameState.getGameStatePieces(), new ArrayList<DraughtPiece>());
		gameState.setTurn(gameState.getNextTurn());
		geneticOpponent.start();
	}



	private void simulateGame(MiniMaxNode rootNode) {
		ChildNodeFinder nodeFinder=new ChildNodeFinder();
		nodeFinder.doAllMoves(rootNode.getGameStatePieces(), playerNum);
		
		//1st level of the tree generated, the state resulting from MiniMaxPlayer's first move
		ArrayList<ArrayList<DraughtPiece>> childStates=nodeFinder.getChildStates();
		
		for(ArrayList<DraughtPiece> childState:childStates) {
			Double miniMaxScore=0.0;
			rootNode.addChild(new MiniMaxNode(childState, miniMaxScore, "minimiser"));
		}
		
		//2nd level of the tree generated- this is the state resulting from the opponent's move
		ArrayList<MiniMaxNode> levelOneNodes=rootNode.getChildren();
		for(MiniMaxNode levelOneNode:levelOneNodes) {
			ArrayList<DraughtPiece> levelOneState=levelOneNode.getGameStatePieces();
			nodeFinder.doAllMoves(levelOneState, opponentNum);
			
			ArrayList<ArrayList<DraughtPiece>> levelTwoStates=nodeFinder.getChildStates();
			for(ArrayList<DraughtPiece> levelTwoState:levelTwoStates) {
				Double miniMaxScore=0.0;
				MiniMaxNode levelTwoNode=new MiniMaxNode(levelTwoState, miniMaxScore, "maximiser");
				levelOneNode.addChild(levelTwoNode);
			}
		}
		
		//Third level of the tree generated, the state resulting from MiniMaxPlayer's second move
		for(MiniMaxNode levelOneNode:levelOneNodes) {
			ArrayList<MiniMaxNode> levelTwoNodes=levelOneNode.getChildren();
			
			for(MiniMaxNode levelTwoNode:levelTwoNodes) {
				ArrayList<DraughtPiece> levelTwoState=levelTwoNode.getGameStatePieces();
				nodeFinder.doAllMoves(levelTwoState, playerNum);
				
				ArrayList<ArrayList<DraughtPiece>> levelThreeStates=nodeFinder.getChildStates();
				for(ArrayList<DraughtPiece> levelThreeState:levelThreeStates) {
					Double miniMaxScore=evaluateState(levelThreeState);
					MiniMaxNode levelThreeNode=new MiniMaxNode(levelThreeState, miniMaxScore, "minimiser");
					levelTwoNode.addChild(levelThreeNode);
				}
			}
		}
	}

	private Double evaluateState(ArrayList<DraughtPiece> leafState) {
		Double miniMaxScore=0.0;
		Double leafWhiteMen=0.0, leafRedMen=0.0, leafWhiteKings=0.0, leafRedKings=0.0;
		
		for(DraughtPiece piece:leafState) {
			if(piece.getColor()==Color.WHITE) {
				++leafWhiteMen;
				if(piece.getPieceType()=="king") {
					++leafWhiteKings;
				}
			}
			
			if(piece.getColor()==Color.RED) {
				++leafRedMen;
				if(piece.getPieceType()=="king") {
					++leafRedKings;
				}
			}
		}
		
		int ourManWeight, opponentManWeight, ourKingWeight, opponentKingWeight;
		ourManWeight=heuristicWeights[0];
		opponentManWeight=heuristicWeights[1];
		ourKingWeight=heuristicWeights[2];
		opponentKingWeight=heuristicWeights[3];
	
		if(playerNum==1) {
			//We are the white player here
			miniMaxScore=(ourManWeight*leafWhiteMen)-(opponentManWeight*leafRedMen); //Weights the difference in pawns
			miniMaxScore=miniMaxScore+(ourManWeight*leafWhiteKings)-(opponentManWeight*leafRedKings); //Weights the difference in Kings
			
		} else {
			//We are the red player here
			miniMaxScore=(ourManWeight*leafRedMen)-(opponentManWeight*leafWhiteMen); //Weights the difference in pawns
			miniMaxScore=miniMaxScore+(ourManWeight*leafRedKings)-(opponentManWeight*leafWhiteKings); //Weights the difference in Kings
			
		}
	
		
		//Holding the centre of the board is regarded as desirable, due to the versatility of further movement it provides,
		//and the lesser distance to every other point on the board. Thus this weighting will increase a future state's 
		//score if that state has more pieces centrally located, by seeking to minimise the distance of all pieces
		//to the centre board square. 
		int piecesInCentre=0;
		int totalControlledPieces=0;
		int piecesInCentreWeight=heuristicWeights[4];
		
		if(playerNum==1) {
			//White player here
			for(DraughtPiece piece:leafState) {
				if(piece.getColor().equals(Color.WHITE)) {
					++totalControlledPieces;
					if(piece.getRow()>=6 && piece.getRow()<=3) {
						if(piece.getCol()>=3 && piece.getCol()<=6) {
							++piecesInCentre;
						}
					}
				}
			}
		}
		
		else {
			//Red player here
			for(DraughtPiece piece:leafState) {
				if(piece.getColor().equals(Color.RED)) {
					++totalControlledPieces;
					if(piece.getRow()>=3 && piece.getRow()<=6) {
						if(piece.getCol()>=3 && piece.getCol()<=6) {
							++piecesInCentre;
						}
					}
				}
			}
		}
		
		//Increase the MiniMax score if a greater portion of our pieces are in the centre of the board. 
		miniMaxScore=miniMaxScore+piecesInCentreWeight*(piecesInCentre/totalControlledPieces);
	
		
		return miniMaxScore;
	}
	
	private Double miniMax(MiniMaxNode inputNode) {
		Double bestScore=0.0;
		
		if(inputNode.getChildren().isEmpty()) {
			bestScore=inputNode.getMiniMaxScore();
		}
		
		else if(inputNode.getPlayerType()=="maximiser") {
			bestScore=Double.NEGATIVE_INFINITY;
			
			for(int i=0; i<inputNode.getChildren().size(); ++i) {
				MiniMaxNode child=inputNode.getChildren().get(i);
				Double childScore=miniMax(child);
				bestScore=Math.max(childScore, bestScore);
				
				inputNode.setMiniMaxScore(bestScore);
			}
		}
		
		else {
			bestScore=Double.POSITIVE_INFINITY;
			
			for(int i=0; i<inputNode.getChildren().size(); ++i) {
				MiniMaxNode child=inputNode.getChildren().get(i);
				Double childScore=miniMax(child);
				bestScore=Math.min(childScore, bestScore);
				
				inputNode.setMiniMaxScore(bestScore);
			}
		}
		return bestScore;
	}



	public void setOpponent(GeneticPlayer player1) {
		geneticOpponent=player1;
	}



	public void setOpponent(RandomPlayer player1) {
		// TODO Auto-generated method stub
		
	}

	

}
