package players;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Random;

import board.Board;
import board.DraughtPiece;
import gameControl.GameState;
import genetic.GeneticPlayer;
import moveFinders.LegalPieceFinder;
import tree.ChildNodeFinder;
import tree.MiniMaxNode;

public class BestFirstPlayer {
	private int playerNum;
	private int opponentNum;
	private GameState gameState;
	private Board board;
	private MiniMaxPlayer miniMaxOpponent;
	private HumanPlayer humanOpponent;
	private RandomPlayer randomOpponent;
	private GeneticPlayer geneticOpponent;
	private String opponentType;
	
	private int[] weights;
	
	public BestFirstPlayer(Board board, int playerNum, GameState gameState, int[] weights) {
		this.playerNum=playerNum;
		this.gameState=gameState;
		this.board=board;
		if(playerNum==1) {
			opponentNum=2;
		} else {
			opponentNum=1;
		}
		this.weights=weights;
	}
	
	public void start() {	
		LegalPieceFinder legalPieceFinder=new LegalPieceFinder();
		ArrayList<DraughtPiece> legalPieces=legalPieceFinder.findLegalPieces(playerNum, gameState.getGameStatePieces());
		
		if(!legalPieces.isEmpty()) {
			MiniMaxNode rootNode=new MiniMaxNode(gameState.getGameStatePieces(), 0.0, "maximiser");
			
			simulateGame(rootNode);
			
			doMove(rootNode);		
		}
		
		else {

		}
	}
	

	private void doMove(MiniMaxNode rootNode) {
		ArrayList<MiniMaxNode> childNodes=rootNode.getChildren();
		
		double bestStateScore=0.0;
		MiniMaxNode chosenNode=null;
		for(MiniMaxNode childNode:childNodes) {
			if(childNode.getMiniMaxScore()>=bestStateScore) {
				bestStateScore=childNode.getMiniMaxScore();
				chosenNode=childNode;
			}
		}
		if(chosenNode==null) {
			Random randomGenerator=new Random();
			int randInt=randomGenerator.nextInt(rootNode.getChildren().size());
			
			chosenNode=rootNode.getChildren().get(randInt);
		}
		
		ArrayList<DraughtPiece> chosenMove=chosenNode.getGameStatePieces();
		
		gameState.setGameState(chosenMove);
		board.repaintBoard(null, gameState.getGameStatePieces(), new ArrayList<DraughtPiece>());				
		
	
		end();
	}

	private void simulateGame(MiniMaxNode rootNode) {
		ChildNodeFinder nodeFinder=new ChildNodeFinder();
		nodeFinder.doAllMoves(rootNode.getGameStatePieces(), playerNum);
		
		//1st level of the tree generated, the state resulting from BestFirstPlayer's first move
		ArrayList<ArrayList<DraughtPiece>> childStates=nodeFinder.getChildStates();
		
		for(ArrayList<DraughtPiece> childState:childStates) {
			Double miniMaxScore=evaluateState(childState);
			rootNode.addChild(new MiniMaxNode(childState, miniMaxScore, "minimiser"));
		}
	}
	
	private Double evaluateState(ArrayList<DraughtPiece> leafState) {
		Double miniMaxScore=0.0;
		Double leafWhiteMen=0.0, leafRedMen=0.0, leafWhiteKings=0.0, leafRedKings=0.0;
		
		for(DraughtPiece piece:leafState) {
			if(piece.getColor()==Color.WHITE) {
				++leafWhiteMen;
				if(piece.getPieceType()=="king") {
					++leafWhiteKings;
				}
			}
			
			if(piece.getColor()==Color.RED) {
				++leafRedMen;
				if(piece.getPieceType()=="king") {
					++leafRedKings;
				}
			}
		}
		
		int manDifferenceWeight=weights[0];
		int kingDifferenceWeight=weights[1];
		
		if(playerNum==1) {
			//We are the white player here
			miniMaxScore=manDifferenceWeight*(leafWhiteMen-leafRedMen); //Weights the difference in men
			miniMaxScore=miniMaxScore+kingDifferenceWeight*(leafWhiteKings-leafRedKings); //Weights the difference in Kings
			
		} else {
			//We are the red player here
			miniMaxScore=manDifferenceWeight*(leafRedMen-leafWhiteMen); //Weights the difference in men
			miniMaxScore=miniMaxScore+kingDifferenceWeight*(leafRedKings-leafWhiteKings); //Weights the difference in Kings;
		}
		
		
		//Holding the centre of the board is regarded as desirable, due to the versatility of further movement it provides,
		//and the lesser distance to every other point on the board. Thus this weighting will increase a future state's 
		//score if that state has more pieces centrally located, by seeking to minimise the distance of all pieces
		//to the centre board square. 
		int piecesInCentre=0;
		int totalControlledPieces=0;
		int piecesInCentreWeight=weights[2];
			
		if(playerNum==1) {
			//White player here
			for(DraughtPiece piece:leafState) {
				if(piece.getColor().equals(Color.WHITE)) {
					++totalControlledPieces;
					if(piece.getRow()>=6 && piece.getRow()<=3) {
						if(piece.getCol()>=3 && piece.getCol()<=6) {
							++piecesInCentre;
						}
					}
				}
			}
		}
				
		else {
			//Red player here
			for(DraughtPiece piece:leafState) {
				if(piece.getColor().equals(Color.RED)) {
					++totalControlledPieces;
					if(piece.getRow()>=3 && piece.getRow()<=6) {
						if(piece.getCol()>=3 && piece.getCol()<=6) {
							++piecesInCentre;
						}
					}
				}
			}
		}
				
		//Increase the MiniMax score if a greater portion of our pieces are in the centre of the board. 
		miniMaxScore=miniMaxScore+piecesInCentreWeight*(piecesInCentre/totalControlledPieces);
				
	
		return miniMaxScore;
	}

	private void end() {
		gameState.setTurn(gameState.getNextTurn());
		if(opponentType=="miniMax") {
			miniMaxOpponent.start();
		} 
		if(opponentType=="random") {
			randomOpponent.start();
		} 
		
		if(opponentType=="human") {
			humanOpponent.start();
		}
		
		if(opponentType=="genetic") {
			geneticOpponent.start();
		}
	}

	public void setOpponent(HumanPlayer player2) {
		humanOpponent=player2;
		opponentType="human";
	}
	
	public void setOpponent(RandomPlayer player2) {
		randomOpponent=player2;
		opponentType="random";
	}
	
	public void setOpponent(GeneticPlayer player2) {
		geneticOpponent=player2;
		opponentType="genetic";
	}
	
	public void setOpponent(MiniMaxPlayer player2) {
		miniMaxOpponent=player2;
		opponentType="miniMax";
	}




}
