package players;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Arrays;

import board.Board;
import board.DraughtPiece;
import gameControl.GameState;
import genetic.GeneticPlayer;
import moveFinders.LegalMoveFinder;
import moveFinders.LegalPieceFinder;
import moveFinders.MoveFinder;
import moveFinders.TurnCalculator;

public class HumanPlayer {
	private Board board;
	private int playerNum;
	private GameState gameState;
	
	private String opponentType;
	private MiniMaxPlayer miniMaxOpponent;
	private RandomPlayer randomOpponent;
	private HumanPlayer humanOpponent;
	
	private ArrayList<int[]> legalMoves;
	private DraughtPiece clickedPiece;
	private LegalPieceFinder legalPieceFinder;
	private ArrayList<DraughtPiece> legalPieces;
	private GeneticPlayer geneticOpponent;
	private BestFirstPlayer bestFirstOpponent;
	
	public HumanPlayer(Board board, int playerNum, GameState gameState) {
		this.board=board;
		this.playerNum=playerNum;
		this.gameState=gameState;
		legalMoves=new ArrayList<int[]>();
		legalPieceFinder=new LegalPieceFinder();
	}

	public void start() {
		legalMoves=new ArrayList<int[]>();
		legalPieces=legalPieceFinder.findLegalPieces(playerNum, gameState.getGameStatePieces());
		board.repaintBoard(null, null, legalPieces);
		
		MouseDetector mouseDetector=new MouseDetector();
		board.addMouseListener(mouseDetector);
	}
	
	private void doTurn(int[] clickedSquare, MouseDetector mouseDetector) {
		for(DraughtPiece piece:legalPieces) {
			if(piece.getCol()==clickedSquare[0] && piece.getRow()==clickedSquare[1]) {
				clickedPiece=piece;
				
				TurnCalculator turnCalculator=new TurnCalculator();
				legalMoves=turnCalculator.findLegalMoves(clickedPiece, gameState.getGameStatePieces());
				
				board.repaintBoard(legalMoves, gameState.getGameStatePieces(), null);
			}
		}
		
		for(int[] legalMove:legalMoves) {
			if(Arrays.equals(clickedSquare, legalMove)) {
				doMove(mouseDetector, clickedPiece, legalMove);
				break;
			}
		}
	}
	
	private void doMove(MouseDetector mouseDetector, DraughtPiece clickedPiece, int[] chosenMove) {
		MoveFinder moveFinder=new MoveFinder();
		DraughtPiece movedPiece=moveFinder.movePiece(clickedPiece, chosenMove);
		DraughtPiece takenPiece=moveFinder.findTakenPiece(clickedPiece, chosenMove, gameState.getGameStatePieces());
		
		ArrayList<DraughtPiece> oldGameStatePieces=new ArrayList<DraughtPiece>();
		oldGameStatePieces.addAll(gameState.getGameStatePieces());
		ArrayList<DraughtPiece> newGameStatePieces=gameState.getGameStatePieces();
		 		
		newGameStatePieces.remove(clickedPiece);
		newGameStatePieces.remove(takenPiece);
		newGameStatePieces.add(movedPiece);
		
		gameState.setGameState(newGameStatePieces);
		board.repaintBoard(null, gameState.getGameStatePieces(), null);
		
		
		if(moveFinder.pieceTaken(oldGameStatePieces, newGameStatePieces)) {
			if(moveFinder.pieceKinged()) {
				end(mouseDetector);
			} 
			
			else {
				legalPieces=legalPieceFinder.findPossibleJumpPiece(movedPiece, gameState.getGameStatePieces());
				if(legalPieces.isEmpty()) {
					end(mouseDetector);
				} else {
				board.repaintBoard(null, newGameStatePieces, legalPieces);
				}	
			}
			
		} 
		
		else {
			end(mouseDetector);
		}				

	}
	
	
	
	
	
	
	
	
	
	private void end(MouseDetector mouseDetector) {
		board.removeMouseListener(mouseDetector);
		gameState.setTurn(gameState.getNextTurn());
		if(opponentType=="miniMax") {
			miniMaxOpponent.start();
		} 
		if(opponentType=="random") {
			randomOpponent.start();
		} 
		
		if(opponentType=="human") {
			humanOpponent.start();
		}
		
		if(opponentType=="genetic") {
			geneticOpponent.start();
		}
		if(opponentType=="bestFirst") {
			bestFirstOpponent.start();
		}

	}
	
	

	public class MouseDetector extends MouseAdapter{
		public void mouseClicked(MouseEvent e) {
			int[] clickedSquare=findClickedSquare(e.getX(), e.getY());			
			doTurn(clickedSquare, this);
		}
		
		private int[] findClickedSquare(int mouseX, int mouseY) {
			int[] squareCoords=null;
			int originX=195;
			int originY=30;
			for(int row=1; row<9; ++row) {
				for(int col=1; col<9; ++col) {
					Rectangle2D square=new Rectangle2D.Double(originX+(62*col), originY+(62*row), 62, 62);
					if(square.contains(mouseX, mouseY)) {
						squareCoords=new int[2];
						squareCoords[0]=col;
						squareCoords[1]=row;
						break;
					}
				}
			}
			return squareCoords;
		}
	}
	

	
	public void setOpponent(MiniMaxPlayer player2) {
		miniMaxOpponent=player2;
		opponentType="miniMax";
	}
	
	public void setOpponent(HumanPlayer player2) {
		humanOpponent=player2;
		opponentType="human";
	}
	
	public void setOpponent(RandomPlayer player2) {
		randomOpponent=player2;
		opponentType="random";
	}
	
	public void setOpponent(GeneticPlayer player2) {
		geneticOpponent=player2;
		opponentType="genetic";
	}

	public void setOpponent(BestFirstPlayer player2) {
		bestFirstOpponent=player2;
		opponentType="bestFirst";
	}


}	
